/*
Kernel Module to implement rdma_sendfile

Copyright (C) 2007 - Dennis Dalessandro  -  Ohio Supercomputer Center
                                         (dennis@osc.edu)

Distributed under the GNU Public License Version 2  (SEE LICENSE)
*/
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/security.h>
#include <linux/pagemap.h>
#include <linux/device.h>
#include <linux/version.h>
#include <linux/idr.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <rdma/ib_verbs.h>
#include <rdma/ib_fmr_pool.h>
#include <../drivers/infiniband/core/uverbs.h>  /* struct ib_uverbs_file */
#include "rdma_sendfile.h"



MODULE_LICENSE("GPL");
MODULE_AUTHOR("OSC iwarp team");
MODULE_DESCRIPTION("Implementation of RDMA SendFile");

static int major_number;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	#define MODERN_LINUX
	extern struct ib_pd *idr_read_pd(int pd_handle,
		struct ib_ucontext *context);
	extern void put_pd_read(struct ib_pd *pd);
#else
	extern struct idr ib_uverbs_pd_idr;
#endif

/*must get a page and send it across on the socket fd which is still carried
along in desc->arg.data*/
int rsf_file_send_actor(read_descriptor_t *desc, struct page *page,
                        unsigned long offset, unsigned long size)
{
	ssize_t written;
	unsigned long count = desc->count;  /*number of bytes left to read
			     desc.arg.data->count is total count user sked for*/
	rsf_desc_t *rsf_info = desc->arg.data;
	struct ib_phys_buf *phys_buf;
	int access_flags;
	u64 iova = 0;
	unsigned long total_size;
	int ret;
	int i;
	int curr_num = rsf_info->user_info->current_queue_num;
	int page_count;

	#ifdef SKIP_LOOP
		goto SKIP;
	#endif


	rsf_debug(9, "Count is %ld Size is %ld", count, size);
	//~ if (size != PAGE_SIZE){
		//~ printk(KERN_ERR "File must be a multiple of page size "
				//~ "currently!");
		//~ return -EINVAL;
	//~ }

	/*count is total amt left, size is amt in one round
	size might be > count so need to adjust it*/
	if (size > count)
		size = count;

	page_count = rsf_info->page_count[curr_num];

	if (page_count >= MAX_PAGE) {
		printk(KERN_ERR "Can't handle this many pages max is %d\n",
					MAX_PAGE);
		return -ENOMEM;
	}

	/*add the page to the array and skip the rest*/
	rsf_debug(9, "Remembering page %d: %p", page_count, page);
	rsf_info->page[curr_num][page_count] = page;
	rsf_info->page_count[curr_num] = ++page_count;
	page_cache_get(page);   /*get a reference to the page*/

	if (count > PAGE_SIZE)
		goto remember_page_only;  /* still more to do */

	/*now we have an array full of pages that need registered*/
	if (page_count * sizeof(*phys_buf) > PAGE_SIZE)
		return -ENOMEM;
	phys_buf = page_address(rsf_info->phys_buf_page);
	for (i=0; i<page_count; i++) {
		dma_addr_t phys_addr =
			page_to_phys(rsf_info->page[curr_num][i]);
		phys_buf[i].addr = phys_addr;
		phys_buf[i].size = PAGE_SIZE;
		rsf_debug(9, "Adding page info to phys buf: %d - 0x%llx - %llu "
					"bytes", i, phys_buf[i].addr,
					phys_buf[i].size);
	}
	rsf_debug(5, "Just added %d pages", page_count);
	if (size < PAGE_SIZE)
		phys_buf[page_count-1].size = size; /* last page may be short */
	iova = phys_buf[0].addr;
	//~ access_flags = IB_ACCESS_LOCAL_WRITE;  /* XXX: should be read-only */
	access_flags = 0;
	rsf_debug(5, "current queue num is %d, page_count is %d",
	          curr_num, page_count);
	rsf_debug(5, "iova is %llx before reg", iova);
	#ifdef NO_REG
	rsf_info->mr[curr_num] = &rsf_info->junk;
	#else
	rsf_info->mr[curr_num] = ib_reg_phys_mr(rsf_info->prot_id, phys_buf,
						page_count, access_flags,
						&iova);
	rsf_debug(5, "iova is %llx after reg", iova);
	if (rsf_info->mr[curr_num] == NULL) {
		printk(KERN_ERR "Unable to register physical buffer");
		return -EINVAL;
	}
	rsf_debug(5, "Buffer is registered, lkey = 0x%x",
				rsf_info->mr[curr_num]->lkey);

	/*sge*/
	ret = copy_to_user(rsf_info->user_info->lkey,
			&rsf_info->mr[curr_num]->lkey, sizeof(uint32_t));

	/*addr*/
	ret = copy_to_user(rsf_info->user_info->addr, &iova, sizeof(uint64_t));

	if (size < PAGE_SIZE)
		total_size = PAGE_SIZE * (page_count - 1) + size;
	else
		total_size = PAGE_SIZE * page_count;
	rsf_debug(5, "Actual Size is %lu", total_size);

	/*actual len*/
	ret = copy_to_user(rsf_info->user_info->actual_len, &total_size,
					sizeof(unsigned long));
	#endif

#ifdef SKIP_LOOP
SKIP:
#endif

remember_page_only:
	written = size;
	desc->count = count - written;
	desc->written += written;
	return written;
}

#define MAX_RW_COUNT (INT_MAX & PAGE_CACHE_MASK)

int rsf_rw_verify_area(int read_write, struct file *file, loff_t *ppos,
			size_t count)
{
	struct inode *inode;
	loff_t pos;

	if (unlikely((ssize_t) count < 0))
		goto Einval;
	pos = *ppos;
	if (unlikely(pos < 0 || pos + count < 0))
		goto Einval;

	inode = file->f_dentry->d_inode;
	if (unlikely(inode->i_flock && MANDATORY_LOCK(inode))) {
		int retval = locks_mandatory_area(
			read_write == READ ? FLOCK_VERIFY_READ :
						FLOCK_VERIFY_WRITE,
			inode, file, pos, count);
		if (retval < 0)
			return retval;
	}
	return count > MAX_RW_COUNT ? MAX_RW_COUNT : count;

Einval:
	return -EINVAL;
}

static ssize_t rsf_do_sendfile(rsf_desc_t *desc)
{
	struct file * in_file;
	struct inode * in_inode;
	ssize_t retval;
	int in_fd;
	loff_t *ppos;
	size_t count;

	in_fd = desc->user_info->fd;
	ppos = &desc->user_info->offset;
	count = desc->user_info->count;

	/*
	 * Get input file, and verify that it is ok..
	 */
	retval = -EBADF;
	in_file = fget(in_fd);
	if (!in_file)
		goto out;
	if (!(in_file->f_mode & FMODE_READ))
		goto fput_in;
	retval = -EINVAL;
	in_inode = in_file->f_dentry->d_inode;
	if (!in_inode)
		goto fput_in;
	if (!in_file->f_op || !in_file->f_op->sendfile)
		goto fput_in;
	retval = -ESPIPE;
	if (!ppos)
		ppos = &in_file->f_pos;
	else
		if (!(in_file->f_mode & FMODE_PREAD))
			goto fput_in;
	retval = rsf_rw_verify_area(READ, in_file, ppos, count);
	if (retval < 0)
		goto fput_in;
	count = retval;

	retval = security_file_permission (in_file, MAY_READ);
	if (retval)
		goto fput_in;


	/*the following should call generic_file_sendfile() in mm/filemap.c
	which calls do_generic_file_read() which calls do_generic_mapping_read()
	which is one ugly function  that in some places calls the actor
	function above */
	retval = in_file->f_op->sendfile(in_file, ppos, count,
				rsf_file_send_actor, desc);
	rsf_debug(9, "Done calling sendfile just set retval to %ld", retval);
	if (retval != count)
		return -EOVERFLOW;
fput_in:
	fput(in_file);
out:
	return retval;
}

static int release_pipe(rsf_desc_t *krsf_info, int slot)
{
	int i, ret;

	rsf_debug(5, "Releasing pages for pipe %d", slot);

	if (slot >= ARRAY_SIZE(krsf_info->mr))
		return -EINVAL;
	#ifndef NO_REG
		#ifndef NO_DEREG
			ret = ib_dereg_mr(krsf_info->mr[slot]);
			if(ret != 0) {
				printk(KERN_ERR "Unable to deregister memory");
				return ret;
			}
		#endif
	#endif
	krsf_info->mr[slot] = NULL;
	for (i=0; i < krsf_info->page_count[slot]; i++) {
		rsf_debug(9, "releasing slot %d page %d: %p", slot, i,
		          krsf_info->page[slot][i]);
		page_cache_release(krsf_info->page[slot][i]);
		krsf_info->page[slot][i] = NULL;
	}
	krsf_info->page_count[slot] = 0;
	return 0;
}

static int rsf_open(struct inode *inode, struct file *file)
{
	rsf_debug(5, "Entry for Open, private data %p", file->private_data);
	file->private_data = NULL;
	return 0;
}

static int rsf_release(struct inode *inode, struct file *file)
{
	int i, ret;
	rsf_desc_t *krsf_info = file->private_data;

	rsf_debug(5, "Entry for Release, info %p", krsf_info);

	if (krsf_info != NULL) {
		/* check the MR array for anything NOT NULL and free pages for
		 * that queue number and then the mr */
		for (i=0; i<MAX_OUTSTANDING; i++) {
			if (krsf_info->mr[i] != NULL) {
				ret = release_pipe(krsf_info, i);
				if (ret)
					return ret;
			}
		}
		__free_pages(krsf_info->phys_buf_page, 0);
		kfree(krsf_info);
	}
	return 0;
}

static ssize_t rsf_read(struct file *filp, char *buffer, size_t length,
			loff_t * offset)
{
	int ret;
	char userbuf[100];

	rsf_debug(5, "Entry for Read");

	ret = copy_to_user(buffer, &userbuf, min(sizeof(userbuf), length));
	if (ret) {
		rsf_debug(5, "Failed to copy back to user");
		return -EFAULT;
	}
	return length;
}

static ssize_t rsf_write(struct file *filp, const char *buff, size_t len,
			loff_t * off)
{
	rsf_desc_t *krsf_info = filp->private_data;
	user_info_t user_info;  /* lives on stack during rsf_do_sendfile */
	int ret;

	if (len != sizeof(user_info))
		return -EINVAL;

	if (copy_from_user(&user_info, buff, len))
		return -EFAULT;

	#ifdef ZERO_WORK
		return len;
	#endif

	/*look at user_info bit mask to see what op we are doing*/
	if (user_info.mask & INIT) {
		rsf_debug(5, "Doing Init, info %p", krsf_info);
		if (krsf_info != NULL)
			return -EINVAL;
		krsf_info = kzalloc(sizeof(*krsf_info), GFP_KERNEL);
		if (!krsf_info)
			return -ENOMEM;
		krsf_info->phys_buf_page = alloc_pages(GFP_KERNEL, 0);
		if (!krsf_info->phys_buf_page)
			return -ENOMEM;
		filp->private_data = krsf_info;
		rsf_debug(5, "Now info %p", krsf_info);
	}

	if (user_info.mask & SEND_DATA) {
		int curr_num;
		int free_num;

		rsf_debug(5, "Doing Send Data, info %p", krsf_info);
		if (krsf_info == NULL) {
			printk(KERN_ERR "Never initialized");
			return -EINVAL;
		}
		krsf_info->user_info = &user_info;

		curr_num = krsf_info->user_info->current_queue_num;
		free_num = krsf_info->user_info->free_queue_num;

		rsf_debug(9, "User wants to send FD %d",
						krsf_info->user_info->fd);
		rsf_debug(9, "User wants to send a total of %d bytes",
						krsf_info->user_info->count);
		rsf_debug(9, "User wants to start at offset %ld", (long int)
						krsf_info->user_info->offset);

		if (free_num >= 0)
			release_pipe(krsf_info, free_num);

		krsf_info->page_count[curr_num] = 0;

		/*using krsf_info.user_info.pd_hndl look up the struct ib_pd
		pointer to save in krsf_info.pd_hndl*/
		#ifdef MODERN_LINUX
			{
			struct file *ib_filp = fget(
						krsf_info->user_info->ib_fd);
			struct ib_uverbs_file *file = ib_filp->private_data;
			krsf_info->prot_id = idr_read_pd(
				krsf_info->user_info->pd_hndl, file->ucontext);
			fput(ib_filp);
			}
		#else
			/* requires mod to kernel source see
			infiniband/core/uverbs_main.c line #77 */
			krsf_info->prot_id = idr_find(&ib_uverbs_pd_idr,
						krsf_info->user_info->pd_hndl);
		#endif
			rsf_debug(5, "PD %p", krsf_info->prot_id);
			ret = rsf_do_sendfile(krsf_info);
			if (ret < 0)
				return ret;

			rsf_debug(5, "%s: done, krsf %p, pd %p, ret %d",
			          __func__, krsf_info, krsf_info->prot_id, ret);
		#ifdef MODERN_LINUX
			put_pd_read(krsf_info->prot_id);
		#endif

		/* debugging: not valid to access after this call */
		krsf_info->user_info = NULL;
	}

	if (user_info.mask & FINI) {
		if (user_info.mask & ~FINI)
			return -EINVAL;  /*FINI has to happen by its self!*/
		rsf_debug(5, "Doing Fini, info %p", krsf_info);
		if (krsf_info == NULL) {
			printk(KERN_ERR "Never initialized");
			return -EINVAL;
		}

		ret = rsf_release(NULL, filp);
		if (ret)
			return ret;
	}

	return len;  /* size "written" with this call */

}

static struct file_operations ops = {
	.owner = THIS_MODULE,
	.read = rsf_read,
	.write = rsf_write,
	.open = rsf_open,
	.release = rsf_release
};

static struct class *rsf_class;
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
static struct device *rsf_device;
#else
static struct class_device *rsf_device;
#endif

static int rsf_init(void)
{
	int ret = 0;

	printk(KERN_INFO "RDMA SendFile Module loaded\n");

	major_number = register_chrdev(0, DEVICE_NAME, &ops);
	if (major_number < 0) {
		rsf_debug(0, "Unable to register char device failed with %d",
					major_number);
		ret = major_number;
		goto out;
	}

	rsf_debug(5, "Assigned major number %d", major_number);
	ret = 0;

	/* make /dev/DEVICE_NAME */
	rsf_class = class_create(THIS_MODULE, DEVICE_NAME);
	if (IS_ERR(rsf_class)) {
		ret = PTR_ERR(rsf_class);
		goto out_chrdev;
	}
	#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
		rsf_device = device_create(rsf_class, NULL,
				MKDEV(major_number,0), DEVICE_NAME);
	#else
		rsf_device = class_device_create(rsf_class, NULL,
				MKDEV(major_number,0), NULL, DEVICE_NAME);
	#endif
	if (IS_ERR(rsf_device)) {
		ret = PTR_ERR(rsf_device);
		goto out_class;
	}

	goto out;

out_class:
	class_destroy(rsf_class);
out_chrdev:
	unregister_chrdev(major_number, DEVICE_NAME);
out:
	return ret;
}

static void rsf_exit(void)
{
	int ret;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,19)
	device_destroy(rsf_class, MKDEV(major_number,0));
#else
	class_device_destroy(rsf_class, MKDEV(major_number,0));
#endif
	class_destroy(rsf_class);

	ret = unregister_chrdev(major_number, DEVICE_NAME);
	if (ret < 0)
		rsf_debug(0, "Unable to unregister character device, error "
					"code: %d\n", ret);

	printk(KERN_INFO "Unloaded Module\n");
}

module_init(rsf_init);
module_exit(rsf_exit);

