/*
Test driver for
Kernel Module to implement rdma_sendfile

Copyright (C) 2007 - Dennis Dalessandro  -  Ohio Supercomputer Center
                                         (dennis@osc.edu)

Distributed under the GNU Public License Version 2  (SEE LICENSE)
*/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <libgen.h>
#include <sys/errno.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/sendfile.h>
#include <stdarg.h>

#include "perfmon.h"
#include "rdma/rdma_cma.h"
#include "infiniband/verbs.h"
//~ #define WRITE_FILE 1
#ifdef PIPE  /*from makefile*/
	#include "pipeline_write.h"
#else
	#include "u_rsf.h"
#endif

#define uint64_from_ptr(p) ((uint64_t)(uintptr_t)(p))
#define IGNORE __attribute__ ((unused))
/*can define one fo the following to generate specialized executable*/
#define DYNAMIC
//~ #define HACK

/*end specialized executable defines*/


#define BLOCKING /*define to make rdma sendfile library sleep*/

typedef enum {
	RDMA = 0,
	TCP = 1,
} transport_mode_t;

typedef enum  {
	WRITE = 0,
	SENDFILE = 1,
} sendfile_mode_t;

typedef struct {  /*software iWARP compatible*/
    uint32_t length;
    uint32_t stag; /*the rkey*/
    uint64_t to;  /*address*/
} dest_sge_t;

typedef struct{
	uint32_t qp_num;
	uint16_t lid;
}remote_info_t;

typedef struct{
	int id;
	transport_mode_t transport_mode;
	sendfile_mode_t sendfile_mode;
	char *filename;
	int conn_fd; /*for TCP*/
	/*for RDMA*/
	struct ibv_qp 		*qp_hndl;
	struct ibv_context 	*ibv_hndl;
	struct rdma_event_channel *rdma_evt_channel;
	struct rdma_cm_id 	*cma_listen_id;
	struct rdma_cm_id   *cma_id;
	struct ibv_cq 		*cq_hndl;
	struct ibv_pd 		*prot_id;
	struct ibv_comp_channel  *cq_channel;
	void *send_buf;
	void *recv_buf;
	void *rdma_buf;
	struct ibv_mr 		*send_mr;
	struct ibv_mr 		*recv_mr;
	struct ibv_mr 		*rdma_mr;
	struct ibv_sge 	send_sge;
	struct ibv_sge		recv_sge;
	struct ibv_sge		rdma_sge;
	struct ibv_send_wr  send_wr;
	struct ibv_send_wr  rdma_wr;
	struct ibv_recv_wr   recv_wr;
	uint32_t			remote_len;
	uint32_t			remote_rkey;
	uint64_t			remote_addr;
	void 			*fast_buf;
	struct ibv_mr 		*dyn_rdma_mr;
	perfmon_timestamp_t	start;
	perfmon_timestamp_t	stop;
}thread_info_t;

char *server;
size_t bufsize = 4096;  /*default to 4K*/
int port = 6789;
int control_size = 1024;

int max_threads = 1;
perfmon_timestamp_t start, stop;
int MY_DEBUG_LEVEL = 1;

int PAGE_THRESHOLD = MAX_PAGE;
int MAX_OPS = MAX_OUTSTANDING;

int count;  /*keep count of no threads connected*/

pthread_mutex_t count_mutex;
pthread_cond_t client_done = PTHREAD_COND_INITIALIZER;
pthread_mutex_t client_done_mutex;
pthread_mutex_t connect_mutex;

#ifdef OF_SOFTWARE
struct ibv_comp_channel *ibv_create_comp_channel( IGNORE struct ibv_context *context){
	return NULL;
}

int ibv_destroy_comp_channel( IGNORE struct ibv_comp_channel *channel){
	return 0;
}

int ibv_get_cq_event( IGNORE struct ibv_comp_channel *channel,
		      IGNORE struct ibv_cq **cq,  IGNORE void **cq_context){
	return 0;
}

void ibv_ack_cq_events( IGNORE struct ibv_cq *cq, IGNORE unsigned int nevents){
	return;
}

#endif


static void sock_read(int fd, char *buffer, int count);
IGNORE static void sock_write(int fd, const char *buffer, int count);
void *wait_connection(void *arg);
static void iwarp_error(const char *msg, int code);
IGNORE static void poll_cq_threaded(thread_info_t *ti);
static void close_threaded_connection(thread_info_t *ti);
static void get_client_info(thread_info_t *ti);
static void do_server(thread_info_t *ti);
static void usage(void);
static void poll_cq_block_threaded(thread_info_t *ti);
static void make_connection(thread_info_t *ti);
void *do_client(void *arg);
static void send_done(thread_info_t *ti);
void print_info(const char *fmt, ...);

#define rsf_out(lvl,fmt,args...) \
    do { \
		if (lvl <= MY_DEBUG_LEVEL) \
	    	print_info(fmt,##args); \
    } while (0)


void print_info(const char *fmt, ...)
{
    va_list ap;
    fprintf(stdout, "[rdmaSendfile]: ");
    va_start(ap, fmt);
    vfprintf(stdout, fmt, ap);
    va_end(ap);
}

static void usage(void)
{
	fprintf(stderr, "Usage: user_test [Options] -x <server> -f <file>\n");
	fprintf(stderr, "       user_test [Options] -c <server>\n");
	fprintf(stderr, "  -s <bufsize> : set buffer size (default %lu)\n",
	        bufsize);
	fprintf(stderr, "  -t : use TCP transport (default RDMA)\n");
	fprintf(stderr, "  -w : use WRITE (default RDMA)\n");
	fprintf(stderr, "  -p : Max number of pages per operation\n");
	fprintf(stderr, "  -o : Max outstanding operations\n");
	fprintf(stderr, "  -z : Number of threads to spawn\n");
	exit(1);
}

static void sock_read(int fd, char *buffer, int count)
{
	int i, offset = 0;
	int num = count;
	rsf_out(5, "reading %d bytes from FD %d into %p", count, fd, buffer);
	while (num > 0) {
		i = read(fd, buffer + offset, num);
		if (i < 0){
			perror("read failed ");
			pthread_exit(0);
		}
		if (i == 0) {
			continue;
		}
		num -= i;
		offset += i;
	}

}

static void sock_write(int fd, const char *buffer, int count)
{
	int i, offset = 0;
	int num = count;
	rsf_out(5, "writing %bytes from %fd into %p", count, fd, buffer);
	while (num > 0) {
		i = write(fd, buffer + offset, num);
		if (i < 0){
		    perror("Write failed ");
		    pthread_exit(0);
		}
		if (i == 0){
			continue;
		}
		num -= i;
		offset += i;

	}
}

static void iwarp_error(const char *msg, int code)
{
	char error_string[256];
	int length_of_error_string = 256;

	strerror_r(code, error_string, length_of_error_string);
    	fprintf(stderr,"----------ERROR----------\n");
	fprintf(stderr, "%s : %s\n", msg, error_string);
	fprintf(stderr,"----------ERROR----------\n");
}

static void poll_cq_threaded(thread_info_t *ti)
{
	int ret;
	struct ibv_wc wc;
	int sleep_time = 1;
	int polls = 10000 ;

	/*poll off the completion queue*/
	for (;;) {
		if (polls == 0) {
			rsf_out(0, "[%d] Nothing showed up in CQ\n", ti->id);
			pthread_exit(0);
		}
		ret = ibv_poll_cq(ti->cq_hndl, 1, &wc);
		if (ret < 0) {
			rsf_out(0,"[%d] Error getting completion event\n", ti->id);
			pthread_exit(0);
		} else if (ret == 0) {
		    if (sleep_time > 0)
			usleep(sleep_time);
		} else {
		    break; /*got a completion event*/
		}
		polls --;
	}

	if (wc.status != IBV_WC_SUCCESS) {
		rsf_out(5, "[%d] Invaild WC status received %d\n", ti->id,
			wc.status);
		pthread_exit(0);
	}
}

static void poll_cq_block_threaded(thread_info_t *ti)
{
	int ret;
	struct ibv_wc wc;
        struct ibv_cq *cq;
        void *cq_context;

	/*check CQ first*/
	ret = ibv_poll_cq(ti->cq_hndl, 1, &wc);
	if (ret < 0) {
		rsf_out(0, "Error getting completion event\n");
		pthread_exit(0);
	} else if (ret == 0) {
		/*block until get a cqe*/
		ret = ibv_get_cq_event(ti->cq_channel, &cq, &cq_context);
		if(ret != 0){
			rsf_out(0, "Failed to get CQ event\n");
			pthread_exit(0);
		}
		ret = ibv_poll_cq(cq, 1, &wc);
		if (ret <= 0) {
			rsf_out(5, "Did not get cq event when expected: %d\n", ret);
			pthread_exit(0);
		}
		ibv_ack_cq_events(cq,1);
	}

	if (wc.status != IBV_WC_SUCCESS) {
		rsf_out(5,"Invaild WC status received %d\n", wc.status);
		pthread_exit(0);
	}
}

static void close_threaded_connection(thread_info_t *ti)
{
	int ret;
	IGNORE struct rdma_cm_event *event;

	/*tear down everything*/
	if (ti->transport_mode == TCP) {
		close(ti->conn_fd);
		pthread_exit(0);
	}

	rsf_out(5, "Getting disconnect event");
	rdma_get_cm_event(ti->rdma_evt_channel, &event);
	if (event->event != RDMA_CM_EVENT_DISCONNECTED) {
		iwarp_error("Unable to get disconnected event", errno);
	} else {
		rsf_out(5, "Got disconnect event");
	}

	ret = rdma_ack_cm_event(event);
	if(ret) {
		iwarp_error("Unable to release event", -ret);
	}

	rsf_out(5, "Got disconnect event and ackd it");

	rdma_destroy_qp(ti->cma_id);


		ret = ibv_dereg_mr(ti->recv_mr);
	if (ret){
		rsf_out(0, "[%d]", ti->id);
		iwarp_error("Unable to deregister recv mem region", ret);
		pthread_exit(0);
	}

	ret = ibv_dereg_mr(ti->send_mr);
	if (ret){
		rsf_out(0, "[%d]", ti->id);
		iwarp_error("Unable to deregister send mem region", ret);
		pthread_exit(0);
	}

	if (ti->rdma_mr != NULL) {
		ret = ibv_dereg_mr(ti->rdma_mr);
		if (ret){
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to deregister rdma mem region", ret);
			pthread_exit(0);
		}
	}


	/*then destory CMA ID*/
	ret = rdma_destroy_id(ti->cma_id);  /*have to do this before tearing down RDMA QP*/
	if(ret){
		iwarp_error("Unable to destroy CMA ID", -ret);
	}

	if(ti->cma_listen_id != NULL){
		ret = rdma_destroy_id(ti->cma_listen_id);
		if(ret){
			iwarp_error("Unable to destroy CMA listen ID", -ret);
		}
	}

	/*finally can tear down PD and CQ*/
	ret = ibv_dealloc_pd(ti->prot_id);
	if(ret){
		iwarp_error("Unable to deallocate Protection domain", ret);
	}

	ret = ibv_destroy_cq(ti->cq_hndl);
	if(ret){
		iwarp_error("Unable to destroy Completion queue", ret);
	}

	/*might as well do this as well*/
	rdma_destroy_event_channel(ti->rdma_evt_channel);





}




static void get_client_info(thread_info_t *ti)
{
	dest_sge_t his_sge;

	if (ti->transport_mode == TCP)
		return;

	#ifdef BLOCKING
	int ret;
	if (ti->cq_channel != NULL) {
		rsf_out(5, "Requesting notification of next post send\n");
		ret = ibv_req_notify_cq(ti->cq_hndl, 0);
		if(ret){
			rsf_out(0, "Unable to request notification on CQ\n");
			pthread_exit(0);
		}
	}
	#endif

	rsf_out(5, "[%d] Getting remote buffer info\n", ti->id);
	poll_cq_block_threaded(ti);
	memcpy(&his_sge, ti->recv_buf, sizeof(his_sge));
	rsf_out(5, "[%d] Received: RDMA Address = 0x%lx RDMA Rkey = %d, Length = %d\n",
		ti->id, his_sge.to, his_sge.stag, his_sge.length);
	ti->remote_addr = his_sge.to;
	ti->remote_rkey = his_sge.stag;
	ti->remote_len = his_sge.length;


}



void *wait_connection(void *arg)
{
	int ret;
	thread_info_t *ti = (thread_info_t *)arg;
	int mode = ti->transport_mode;

	rsf_out(5, "Thread [%d] Active\n", ti->id);

	if (mode == TCP) {  /*then its TCP*/
		struct sockaddr_in incoming_socket;
		socklen_t incoming_s_len = sizeof(incoming_socket);
		int flags=1;

		/*wait for connection*/
		memset(&incoming_socket, 0, incoming_s_len);
		incoming_socket.sin_family = PF_INET;
		incoming_socket.sin_port = htons(port+ti->id);

		/*create socket and set options*/
		ti->conn_fd = socket(PF_INET, SOCK_STREAM, 0);
		if (ti->conn_fd < 0) {
			rsf_out(0, "[%d]", ti->id);
			perror("Unable to create socket");
			pthread_exit(0);
		}
		ret = setsockopt(ti->conn_fd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags)); /*make socket reusable*/
		if (ret < 0) {
			rsf_out(0, "[%d]", ti->id);
			perror("Unable to set socket reusable");
			pthread_exit(0);
		}
		if (bind(ti->conn_fd, (struct sockaddr *)&incoming_socket, incoming_s_len) < 0) {
			rsf_out(0, "[%d]", ti->id);
			perror("Can't bind socket");
			pthread_exit(0);
		}
		rsf_out(5, "[%d] Listening for connection on %d....\n", ti->id,
			port+ti->id);
		if (listen(ti->conn_fd, 1) < 0) {
			printf("[%d]", ti->id);
			perror("Can't set socket listening");
			pthread_exit(0);
		}
		incoming_s_len = sizeof(incoming_socket);
		ti->conn_fd = accept(ti->conn_fd, (struct sockaddr *) &incoming_socket, &incoming_s_len);
		if (ti->conn_fd < 0) {
			printf("[%d]", ti->id);
			perror("Unable to accept connection");
			pthread_exit(0);
		}
	rsf_out(5, "[%d] Socket accepted on %d\n", ti->id, ti->conn_fd);
	} else {
		struct hostent *host;
		struct sockaddr_in sock_addr;
		struct rdma_cm_event *event;
		struct rdma_conn_param conn_param = {
			.responder_resources = 1,
			.initiator_depth = 1,
			.retry_count = 5,
		};
		int cqe_num;
		struct ibv_device_attr device_attr;
		struct ibv_qp_init_attr qp_init_attr;
		struct ibv_recv_wr *bad_wr;

		ti->ibv_hndl = NULL;
		memset(&sock_addr, 0, sizeof(sock_addr));
		sock_addr.sin_family = AF_INET;
		sock_addr.sin_port = htons(port+ti->id);

		ti->rdma_evt_channel = rdma_create_event_channel();

		#if 0
		/*uncomment when neteffect fixes their busted driver*/
		sock_addr.sin_addr.s_addr = INADDR_ANY;
		#else
		host = gethostbyname(server);
		if(host==NULL) {
			rsf_out(5,"[%d]", ti->id);
			iwarp_error("Unable to get host by name", errno);
			pthread_exit(0);
		}
		memcpy((char *) &sock_addr.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
		#endif

		ret = rdma_create_id(ti->rdma_evt_channel, &(ti->cma_listen_id), NULL, RDMA_PS_TCP);
		if (ret) {
			rsf_out(5,"[%d]", ti->id);
			iwarp_error("Unable to create rdma id\n", -ret);
			pthread_exit(0);
		}

		ret = rdma_bind_addr(ti->cma_listen_id, (struct sockaddr *) &sock_addr);
		if (ret) {
		    if (ret == -1){
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("can't bind address\n", errno);
		    } else {
			printf("[%d]", ti->id);
			iwarp_error("can't bind address", -ret);
		    }
		    pthread_exit(0);
		}

		/* 0 means maximum backlog */
		//~ printf("[%d] Listening for connection on %d\n", ti->id,
			//~ ti->id+port);
		ret = rdma_listen(ti->cma_listen_id, 0);
		if (ret) {
			if (ret == -1){
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("Can't listen", errno);
			} else{
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("Can't listen", -ret);
			}
			pthread_exit(0);
		}

		ret = rdma_get_cm_event(ti->rdma_evt_channel, &event);
		if (ret) {
		    if (ret == -1) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to get a CM event", errno);
			pthread_exit(0);
		    } else {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to get a CM event", -ret);
			pthread_exit(0);
		    }
		}

		if (event->event != RDMA_CM_EVENT_CONNECT_REQUEST) {
			rsf_out(0, "[%d]", ti->id);
			rsf_out(0, "inavlid event received, expecting  %d RDMA_CM_EVENT_CONNECT_REQUEST, recevd %d\n", RDMA_CM_EVENT_CONNECT_REQUEST, event->event);
			iwarp_error("Invalid event received", errno);
			pthread_exit(0);
		}

		ti->ibv_hndl = event->id->verbs;  /*FINALLY got this now we can continue*/
		ti->cma_id = event->id;  /*need to remember this for later its where the real connection is*/

		ret = rdma_ack_cm_event(event);
		if (ret) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}

		ret = ibv_query_device(ti->ibv_hndl, &device_attr);
		if (ret < 0) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to query device to figure out max CQs", errno);
			pthread_exit(0);
		}
		cqe_num = device_attr.max_cqe;

		if (cqe_num < 1000) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Device can't handle this many completion entries", errno);
			pthread_exit(0);
		}

		cqe_num = 1000;

		#ifdef BLOCKING
		ti->cq_channel = ibv_create_comp_channel(ti->ibv_hndl);
		#else
		ti->cq_channel = NULL;
		#endif

		ti->cq_hndl = ibv_create_cq(ti->ibv_hndl, cqe_num, NULL, ti->cq_channel, 0);
		if (!ti->cq_hndl) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to create CQ", errno);
			pthread_exit(0);
		}

		ti->prot_id = ibv_alloc_pd(ti->ibv_hndl);
		if (!ti->prot_id) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to create PD", errno);
			pthread_exit(0);
		}


		memset(&qp_init_attr, 0, sizeof(qp_init_attr));
		/* wire both send and recv to the same CQ */
		qp_init_attr.send_cq =  ti->cq_hndl;
		qp_init_attr.recv_cq  = ti->cq_hndl;
		qp_init_attr.cap.max_send_wr = 500;  /* outstanding WQEs */
		qp_init_attr.cap.max_recv_wr = 500;
		qp_init_attr.cap.max_send_sge = 1;  /* scatter/gather entries */
		qp_init_attr.cap.max_recv_sge = 1;
		qp_init_attr.qp_type = IBV_QPT_RC;
		/* only generate completion queue entries if requested */
		qp_init_attr.sq_sig_all = 0;
		ret = rdma_create_qp(ti->cma_id, ti->prot_id, &qp_init_attr);
		if (ret) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("unable to create qp", -ret);
			pthread_exit(0);
		}
		ti->qp_hndl = ti->cma_id->qp;

		ti->send_buf = malloc(control_size);
		ti->recv_buf = malloc(control_size);
		ti->rdma_buf = malloc(bufsize);
		/*register send buffer*/


		ti->send_mr = ibv_reg_mr(ti->prot_id, ti->send_buf, control_size, IBV_ACCESS_LOCAL_WRITE);
		if (!ti->send_mr) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to register send buffer", errno);
			pthread_exit(0);
		}
		 /*register recv buffer*/
		ti->recv_mr = ibv_reg_mr(ti->prot_id, ti->recv_buf, control_size, IBV_ACCESS_LOCAL_WRITE );
		if (!ti->recv_mr) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to register recv buffer", errno);
			pthread_exit(0);
		}
		/*register rdma buffer*/
		ti->rdma_mr = ibv_reg_mr(ti->prot_id, ti->rdma_buf, bufsize,  IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
		if (!ti->rdma_mr) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to register rdma buffer", errno);
			pthread_exit(0);
		}
		memset(ti->send_buf, '\0', control_size);
		memset(ti->rdma_buf, '\0', bufsize);
		memset(ti->recv_buf, '\0', control_size);

		/*Build work requests and post a RECV*/
		/*build recv work request*/
		ti->recv_sge.addr = uint64_from_ptr(ti->recv_buf);
		ti->recv_sge.length = control_size;
		ti->recv_sge.lkey = ti->recv_mr->lkey;

		memset(&(ti->recv_wr), 0, sizeof(ti->recv_wr));
		ti->recv_wr.wr_id = uint64_from_ptr(&ti->recv_wr);
		ti->recv_wr.sg_list = &(ti->recv_sge);
		ti->recv_wr.num_sge = 1;
		ti->recv_wr.next = NULL;



		ti->send_sge.addr = uint64_from_ptr(ti->send_buf);
		ti->send_sge.length = control_size;
		ti->send_sge.lkey = ti->send_mr->lkey;

		memset(&(ti->send_wr), 0, sizeof(ti->send_wr));
		ti->send_wr.wr_id = uint64_from_ptr(&ti->send_wr);
		ti->send_wr.sg_list = &(ti->send_sge);
		ti->send_wr.num_sge = 1;
		ti->send_wr.opcode = IBV_WR_SEND;
		ti->send_wr.send_flags = IBV_SEND_SIGNALED;

		ret = ibv_post_recv(ti->qp_hndl, &ti->recv_wr, &bad_wr);
		if (ret) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Can't post RECV", errno);
			pthread_exit(0);
		}

		/*Finish building the connection*/
		ret = rdma_accept(ti->cma_id, &conn_param);
		if (ret) {
			if (ret == -1){
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("can not accept connection", errno);
			} else {
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("can not accept connection", -ret);
			}
			pthread_exit(0);
		}

		/*now check for connection made event*/
		ret = rdma_get_cm_event(ti->rdma_evt_channel, &event);
		if (ret) {
			if (ret == -1) {
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("Unable to get a CM event", errno);
				pthread_exit(0);
			} else {
				rsf_out(0, "[%d]", ti->id);
				iwarp_error("Unable to get a CM event", -ret);
				pthread_exit(0);
			}
		}
		if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
			rsf_out(0, "[%d]", ti->id);
			rsf_out(0, "inavlid event received, expecting  %d RDMA_CM_EVENT_ESTABLISHED, recevd %d\n", RDMA_CM_EVENT_ESTABLISHED, event->event);
			iwarp_error("Invalid event received", errno);
			pthread_exit(0);
		}

		/*clear connection made event*/
		ret = rdma_ack_cm_event(event);
		if (ret) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}
	}



	rsf_out(5, "Thread %d connected\n", ti->id);

	get_client_info(ti);

	pthread_mutex_lock(&count_mutex);
	count++;
	if (count == max_threads) { /*all threads are connected and ready*/
		rsf_out(5, "All threads connected and ready to go\n");
		pthread_mutex_lock(&client_done_mutex);
		pthread_cond_broadcast(&client_done);
		GET_TIME(start);
		pthread_mutex_unlock(&client_done_mutex);
		count = 0;
		pthread_mutex_unlock(&count_mutex);
	} else {
		pthread_mutex_unlock(&count_mutex);
		pthread_mutex_lock(&client_done_mutex);
		pthread_cond_wait(&client_done, &client_done_mutex);
		pthread_mutex_unlock(&client_done_mutex);
	}


	/*do sendfile*/
	do_server(ti);

	pthread_mutex_lock(&count_mutex);
	count++;
	if (count == max_threads) {
		GET_TIME(stop);
		rsf_out(1, "All threads finished\n");
		pthread_mutex_lock(&client_done_mutex);
		pthread_cond_broadcast(&client_done);
		pthread_mutex_unlock(&client_done_mutex);
		count = 0;
		pthread_mutex_unlock(&count_mutex);
	} else {
		pthread_mutex_unlock(&count_mutex);
		pthread_mutex_lock(&client_done_mutex);
		pthread_cond_wait(&client_done, &client_done_mutex);
		pthread_mutex_unlock(&client_done_mutex);
	}


	pthread_mutex_lock(&client_done_mutex);

	send_done(ti);

	close_threaded_connection(ti);

	pthread_mutex_unlock(&client_done_mutex);

	pthread_exit(0);

}

static void do_server(thread_info_t *ti)
{
	void *file_buff;
	int fd;
	int ret;
	//~ dest_sge_t his_sge;
	transport_mode_t transport_mode = ti->transport_mode;
	sendfile_mode_t sendfile_mode = ti->sendfile_mode;
	char *file_name = ti->filename;


	if (!file_name)
		usage();
	rsf_out(5, "Going to send file %s\n", file_name);

	/*open the file we are going to send to get the FD*/
	ret = open(file_name, O_RDWR);

	if (ret < 0) {
		perror("Can't open file to send");
		pthread_exit(0);
	}
	rsf_out(5, "Opened file to %d\n", ret);
	fd = ret;

	if (sendfile_mode == WRITE) {
		if (transport_mode == TCP) {
			rsf_out(5, "Using usual write() method\n");
			file_buff = malloc(bufsize);
			GET_TIME(ti->start);
			//~ file_buff = mmap(0, bufsize, PROT_READ | PROT_WRITE, MAP_SHARED, conn_fd, 0);
			sock_read(fd, file_buff, bufsize);
			ret = write(ti->conn_fd, file_buff, bufsize);
			GET_TIME(ti->stop);

			rsf_out(9, "TOTAL_WRITE_COST %d %f (us)\n", (int)bufsize, elapsed_wall_time(ti->start, ti->stop, MICROSECONDS));

			if (ret != (ssize_t) bufsize) {
				rsf_out(0, "Did not fully write file\n");
				pthread_exit(0);
			}

			rsf_out(5, "Success\n");
		} else {
			struct ibv_send_wr *bad_wr;

			printf("Using RDMA write() method\n");

			if (bufsize != ti->remote_len) {
				rsf_out(0, "Client and server do not agree on size of RDMA buffer %d %d\n", (int)bufsize, ti->remote_len);
				pthread_exit(0);
			}

			GET_TIME(ti->start);

			#ifdef DYNAMIC
				ti->fast_buf = mmap(0, bufsize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, 0);
				if(ti->fast_buf == (void *) -1){ /*mem map failed*/
					perror("Map of file failed with");
					pthread_exit(0);
				}
				ti->dyn_rdma_mr = ibv_reg_mr(ti->prot_id, ti->fast_buf, bufsize, 0);
				if (!ti->dyn_rdma_mr){
					rsf_out(0, "Unable to register the rdma buffer\n");
					pthread_exit(0);
				}
				ti->rdma_sge.addr = uint64_from_ptr(ti->fast_buf);
				ti->rdma_sge.lkey = ti->dyn_rdma_mr->lkey;
			#else
				ti->rdma_sge.addr = uint64_from_ptr(rdma_buf);
				ti->rdma_sge.lkey = rdma_mr->lkey;
			#endif
			/*Send RDMA buffer*/
			ti->rdma_sge.length = bufsize;

			memset(&ti->rdma_wr, 0, sizeof(ti->rdma_wr));
			ti->rdma_wr.wr_id = uint64_from_ptr(&ti->rdma_wr);
			ti->rdma_wr.sg_list = &ti->rdma_sge;
			ti->rdma_wr.num_sge = 1;
			ti->rdma_wr.wr.rdma.remote_addr = ti->remote_addr;
			ti->rdma_wr.wr.rdma.rkey = ti->remote_rkey;
			ti->rdma_wr.opcode = IBV_WR_RDMA_WRITE;
			ti->rdma_wr.send_flags = IBV_SEND_SIGNALED;  /*set this to 0 to send unsignaled*/

			#ifdef BLOCKING
			if (ti->cq_channel != NULL) {
				rsf_out(5, "Requesting notification of next post send\n");
				ret = ibv_req_notify_cq(ti->cq_hndl, 0);
				if(ret){
					rsf_out(0, "Unable to request notification on CQ\n");
					pthread_exit(0);
				}
			}
			#endif


			/*Copy the file contents into the rdma buffer*/
			#ifndef DYNAMIC
				#ifndef HACK
				sock_read(fd, rdma_buf, bufsize);
				#endif
			#endif
			ret = ibv_post_send(ti->qp_hndl, &ti->rdma_wr, &bad_wr);
			if (ret) {
				iwarp_error("Unable to post to SQ", ret);
				pthread_exit(0);
			}
			#ifdef BLOCKING
				poll_cq_block_threaded(ti);
			#else
				poll_cq_threaded(ti);  /*if we sent it as signaled need to poll it off*/
			#endif

			#ifdef DYNAMIC
			ret = ibv_dereg_mr(ti->dyn_rdma_mr);
			if (ret){
				iwarp_error("Unable to deregister recv mem region", ret);
				pthread_exit(0);
			}
			#endif
			GET_TIME(ti->stop);

			rsf_out(9, "TOTAL_WRITE_COST %d %f (us)\n", (int)bufsize, elapsed_wall_time(ti->start, ti->stop, MICROSECONDS));

			rsf_out(5, "Sent File and Ack\n");

		}
	} else {
		rsf_desc_t rsf_descriptor;

		rsf_out(5, "Using sendfile() method\n");

		if (transport_mode == TCP){
			/*use kernel sendfile directly*/
			GET_TIME(ti->start);
			ret = sendfile(ti->conn_fd, fd, 0, bufsize);
			GET_TIME(ti->stop);
			if(ret != (ssize_t) bufsize){
				perror("Unable to complete sendfile operation");
				pthread_exit(0);
			}
			rsf_out(9, "TOTAL_WRITE_COST %d %f (us)\n", (int)bufsize, elapsed_wall_time(ti->start, ti->stop, MICROSECONDS));
		} else{  /*iwarp or infiniband*/

			ret = u_rsf_init(&rsf_descriptor, ti->qp_hndl, ti->cq_hndl, ti->cq_channel, ti->remote_rkey, ti->remote_addr);
			if(ret != 0){
				rsf_out(0, "Failed to link QP with u_rsf data type\n");
				pthread_exit(0);
			}
			rsf_descriptor.ib_fd = ti->ibv_hndl->cmd_fd;

			if (bufsize != ti->remote_len) {
				iwarp_error("Client and server do not agree on size of RDMA buffer", 0);
				pthread_exit(0);
			}

			ret = u_rsf_modify_threshold(&rsf_descriptor, PAGE_THRESHOLD);
			if(ret != 0){
				rsf_out(0, "Unable to modify threshold\n");
				pthread_exit(0);
			}

			ret = u_rsf_modify_outstanding(&rsf_descriptor, MAX_OPS);
			if(ret != 0){
				rsf_out(0, "Unable to modify outsanding ops\n");
				pthread_exit(0);
			}

			#ifdef SEND_BOGUS
				rsf_descriptor.rdma_buf = ti->rdma_buf;
				rsf_descriptor.addr = uint64_from_ptr(ti->rdma_buf);
				rsf_descriptor.stag = ti->rdma_mr->lkey;
			#endif

			GET_TIME(ti->start);
			#ifdef PIPELINE
				rsf_descriptor.rdma_buf = ti->rdma_buf;
				rsf_descriptor.addr = uint64_from_ptr(ti->rdma_buf);
				rsf_descriptor.stag = ti->rdma_mr->lkey;
			#endif
			ret = rdma_sendfile(fd, &rsf_descriptor, 0, bufsize);
			if (ret != (ssize_t) bufsize) {
				perror("Unable to complete SENDFILE operation\n");
				pthread_exit(0);
			}

			ret = u_rsf_fini(&rsf_descriptor);
			if (ret !=0 ){
				iwarp_error("Unable to fini", ret);
				pthread_exit(0);
			}
			GET_TIME(ti->stop);
			rsf_out(9, "TOTAL_WRITE_COST %d %f (us)\n", (int)bufsize, elapsed_wall_time(ti->start, ti->stop, MICROSECONDS));



			rsf_out(5, "Sent File and Ack\n");



		}

	}
}




static void make_connection(thread_info_t *ti)
{
	int ret;
	transport_mode_t mode = ti->transport_mode;

	if (mode == TCP) {

		struct hostent *host;
		struct sockaddr_in server_address, local_address;
		int connected = 0;
		int k;
		int flags=1;

		rsf_out(5, "[%d] Making connection to server %s\n", ti->id, server);
		host = gethostbyname(server);
		if (host==NULL) {
			perror("Unable to get host by name");
			pthread_exit(0);
		}
		server_address.sin_family = host->h_addrtype;
		memcpy((char *) &server_address.sin_addr.s_addr, host->h_addr_list[0], host->h_length);
		server_address.sin_port = htons(port+ti->id);

		/*create the socket*/
		ti->conn_fd = socket(PF_INET, SOCK_STREAM, 0);
		if (ti->conn_fd < 0) {
			perror("Unable to create socket FD");
			pthread_exit(0);
		}
		/*set socket opts and then bind a port so we can talk with the someone*/
		local_address.sin_family = PF_INET;
		local_address.sin_addr.s_addr = htonl(INADDR_ANY);
		local_address.sin_port = htons(0);
		ret = setsockopt(ti->conn_fd, SOL_SOCKET, SO_REUSEADDR, &flags, sizeof(flags));
		if (ret < 0) {
			perror("Unable to set socket to reusable");
		}

		/*set socket to non blocking since we don't know how big the file incoming is*/
		flags |= O_NONBLOCK;
		ret = fcntl(ti->conn_fd, F_SETFL, O_NONBLOCK);
		if (ret != 0) {
			perror("Can't set socket to non blocking");
			pthread_exit(0);
		}

		ret = bind(ti->conn_fd, (struct sockaddr *) &local_address, sizeof(local_address));
		if (ret < 0) {
			perror("Unable to bind socket");
			pthread_exit(0);
		}
		for (k=0; k<100000; k++) {
			ret = connect(ti->conn_fd, (struct sockaddr *)&server_address, sizeof(server_address));
			if (ret<0) {
				connected = 0;
				usleep(1);
			} else {
				connected = 1;
				break;
			}
		}
		if (connected == 0) {
			perror("Unable to connect to server");
			pthread_exit(0);
		}

		rsf_out(5, "[%d]  Successfully connected to server\n", ti->id);


	} else {

		struct hostent *host;
		struct sockaddr_in sock_addr;
		struct rdma_cm_event *event;
		struct rdma_conn_param conn_param = {
			.responder_resources = 1,
			.initiator_depth = 1,
			.retry_count = 5,
		};
		int cqe_num;
		struct ibv_device_attr device_attr;
		struct ibv_qp_init_attr qp_init_attr;
		struct ibv_recv_wr *bad_wr;

		rsf_out(5, "[%d] in make connection \n", ti->id);

		ti->ibv_hndl = NULL;

		memset(&sock_addr, 0, sizeof(sock_addr));
		sock_addr.sin_family = AF_INET;
		sock_addr.sin_port = htons(port+ti->id);

		ti->rdma_evt_channel = rdma_create_event_channel();
		host = gethostbyname(server);
		if (!host) {
			iwarp_error("Unable to resolve host", errno);
			pthread_exit(0);
		}
		memcpy(&sock_addr.sin_addr, host->h_addr_list[0], host->h_length);
		#ifdef OLD_STACK
		ret = rdma_create_id(ti->rdma_evt_channel, &ti->cma_id, NULL);
		#else
		ret = rdma_create_id(ti->rdma_evt_channel, &ti->cma_id, NULL, RDMA_PS_TCP);
		#endif
		if (ret) {
			iwarp_error("Unable to create RDMA ID", -ret);
			pthread_exit(0);
		}

		rsf_out(5, "[%d] Created event channel, now resolving addr\n", ti->id);

		ret = rdma_resolve_addr(ti->cma_id, NULL, (struct sockaddr *) &sock_addr, 3000);  /* 3 sec timeout */
		if (ret) {
		    iwarp_error("Unable to resolve address", -ret);
		    pthread_exit(0);
		}

		ret = rdma_get_cm_event(ti->rdma_evt_channel, &event);
		if (ret) {
			if (ret == -1) {
				iwarp_error("Unable to get a CM event", errno);
				pthread_exit(0);
			} else {
				iwarp_error("Unable to get a CM event", -ret);
				pthread_exit(0);
			}
		}

		if (event->event != RDMA_CM_EVENT_ADDR_RESOLVED) {
			rsf_out(0, "inavlid event received, expecting  %d RDMA_CM_EVENT_ADDR_RESOLVED, recevd %d\n", RDMA_CM_EVENT_ADDR_RESOLVED, event->event);
			if (event->event == RDMA_CM_EVENT_ADDR_ERROR)
				rsf_out(0, "The invalid event received is RDMA_CM_EVENT_ADDR_ERROR\n");
			iwarp_error("Invalid event received", errno);
			ret = rdma_ack_cm_event(event);
			if (ret)
				iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}

		ti->ibv_hndl = event->id->verbs;  /*FINALLY got this now we can continue*/

		ret = rdma_ack_cm_event(event);
		if (ret) {
			iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}
		rsf_out(5, "[%d] Addr resolved querying dev to make cqs\n", ti->id);
		ret = ibv_query_device(ti->ibv_hndl, &device_attr);
		if (ret < 0) {
			iwarp_error("Unable to query device to figure out max CQs", errno);
			pthread_exit(0);
		}
		cqe_num = device_attr.max_cqe;

		if (cqe_num < 1000) {
			iwarp_error("Device can't handle this many completion entries", errno);
			pthread_exit(0);
		}
		cqe_num = 1000;


		#ifdef BLOCKING
		ti->cq_channel = ibv_create_comp_channel(ti->ibv_hndl);
		#else
		ti->cq_channel = NULL;
		#endif

		ti->cq_hndl = ibv_create_cq(ti->ibv_hndl, cqe_num, NULL, ti->cq_channel, 0);
		if (!ti->cq_hndl) {
			rsf_out(0, "[%d]", ti->id);
			iwarp_error("Unable to create CQ", errno);
			pthread_exit(0);
		}

		//~ ti->cq_hndl = ibv_create_cq(ti->ibv_hndl, cqe_num, NULL, NULL, 0);
		//~ if (!ti->cq_hndl) {
			//~ iwarp_error("Unable to create CQ", errno);
			//~ pthread_exit(0);
		//~ }

		ti->prot_id = ibv_alloc_pd(ti->ibv_hndl);
		if (!ti->prot_id) {
			iwarp_error("Unable to create PD", errno);
			pthread_exit(0);
		}

		rsf_out(5, "[%d] CQs and PDs made now making QP\n", ti->id);
		memset(&qp_init_attr, 0, sizeof(qp_init_attr));
		/* wire both send and recv to the same CQ */
		qp_init_attr.send_cq =  ti->cq_hndl;
		qp_init_attr.recv_cq  = ti->cq_hndl;
		qp_init_attr.cap.max_send_wr = 200;  /* outstanding WQEs */
		qp_init_attr.cap.max_recv_wr = 200;
		qp_init_attr.cap.max_send_sge = 1;  /* scatter/gather entries */
		qp_init_attr.cap.max_recv_sge = 1;
		qp_init_attr.qp_type = IBV_QPT_RC;
		/* only generate completion queue entries if requested */
		qp_init_attr.sq_sig_all = 0;
		ret = rdma_create_qp(ti->cma_id, ti->prot_id, &qp_init_attr);
		if (ret) {
			iwarp_error("unable to create qp", -ret);
			pthread_exit(0);
		}
		ti->qp_hndl = ti->cma_id->qp;

		ti->send_buf = malloc(control_size);
		ti->recv_buf = malloc(control_size);
		ti->rdma_buf = malloc(bufsize);
		/*register send buffer*/
		ti->send_mr = ibv_reg_mr(ti->prot_id, ti->send_buf, control_size, IBV_ACCESS_LOCAL_WRITE);
		if (!ti->send_mr) {
			iwarp_error("Unable to register send buffer", errno);
			pthread_exit(0);
		}
		 /*register recv buffer*/
		ti->recv_mr = ibv_reg_mr(ti->prot_id, ti->recv_buf, control_size, IBV_ACCESS_LOCAL_WRITE );
		if (!ti->recv_mr) {
			iwarp_error("Unable to register recv buffer", errno);
			pthread_exit(0);
		}
		/*register rdma buffer*/
		ti->rdma_mr = ibv_reg_mr(ti->prot_id, ti->rdma_buf, bufsize,  IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE);
		if (!ti->rdma_mr) {
			iwarp_error("Unable to register rdma buffer", errno);
			pthread_exit(0);
		}
		memset(ti->send_buf, '\0', control_size);
		memset(ti->rdma_buf, '\0', bufsize);
		memset(ti->recv_buf, '\0', control_size);

		/*Build work requests and post a RECV*/
		/*build recv work request*/
		ti->recv_sge.addr = uint64_from_ptr(ti->recv_buf);
		ti->recv_sge.length = control_size;
		ti->recv_sge.lkey = ti->recv_mr->lkey;

		memset(&(ti->recv_wr), 0, sizeof(ti->recv_wr));
		ti->recv_wr.wr_id = uint64_from_ptr(&ti->recv_wr);
		ti->recv_wr.sg_list = &(ti->recv_sge);
		ti->recv_wr.num_sge = 1;
		ti->recv_wr.next = NULL;

		ti->send_sge.addr = uint64_from_ptr(ti->send_buf);
		ti->send_sge.length = control_size;
		ti->send_sge.lkey = ti->send_mr->lkey;

		memset(&(ti->send_wr), 0, sizeof(ti->send_wr));
		ti->send_wr.wr_id = uint64_from_ptr(&ti->send_wr);
		ti->send_wr.sg_list = &(ti->send_sge);
		ti->send_wr.num_sge = 1;
		ti->send_wr.opcode = IBV_WR_SEND;
		ti->send_wr.send_flags = IBV_SEND_SIGNALED;

		ret = ibv_post_recv(ti->qp_hndl, &ti->recv_wr, &bad_wr);
		if (ret) {
			iwarp_error("Can't post RECV", errno);
			pthread_exit(0);
		}
		rsf_out(5, "[%d] Resolving route to server\n", ti->id);
		/*now we can resolve the route to the server*/
		ret = rdma_resolve_route(ti->cma_id, 30000);
		if (ret) {
			iwarp_error("Unable to resolve route", -ret);
			pthread_exit(0);
		}

		/*get the route resolved event*/
		ret = rdma_get_cm_event(ti->rdma_evt_channel, &event);
		if (ret) {
			if (ret == -1) {
				iwarp_error("Unable to get a CM event", errno);
				pthread_exit(0);
			} else {
				iwarp_error("Unable to get a CM event", -ret);
				pthread_exit(0);
			}
		}

		if (event->event != RDMA_CM_EVENT_ROUTE_RESOLVED) {
			rsf_out(0, "inavlid event received, expecting  %d RDMA_CM_EVENT_ROUTE_RESOLVED, recevd %d\n", RDMA_CM_EVENT_ROUTE_RESOLVED, event->event);
			iwarp_error("Invalid event received", errno);
			pthread_exit(0);
		}

		/*clear route resolved event*/
		ret = rdma_ack_cm_event(event);
		if (ret) {
			iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}
TRYAGAIN:
		rsf_out(5, "[%d] Making connection\n", ti->id);
		/*attempt to make the connection*/
		ret = rdma_connect(ti->cma_id, &conn_param);
		if (ret) {
			iwarp_error("Unable to connect", -ret);
			pthread_exit(0);
		}

		/*now check for connection made event*/
		ret = rdma_get_cm_event(ti->rdma_evt_channel, &event);
		if (ret) {
			if (ret == -1) {
				iwarp_error("Unable to get a CM event", errno);
				pthread_exit(0);
			} else {
				iwarp_error("Unable to get a CM event", -ret);
				pthread_exit(0);
			}
		}

		if (event->event != RDMA_CM_EVENT_ESTABLISHED) {
			rsf_out(0, "inavlid event received, expecting  %d RDMA_CM_EVENT_ESTABLISHED, recevd %d\n", RDMA_CM_EVENT_ESTABLISHED, event->event);
			iwarp_error("Invalid event received", errno);
			goto TRYAGAIN;
			//~ pthread_exit(0);
		}
		rsf_out(5, "[%d] Done connecting\n", ti->id);
		/*clear connection made event*/
		ret = rdma_ack_cm_event(event);
		if (ret) {
			iwarp_error("Unable to release event", -ret);
			pthread_exit(0);
		}

	}


}


 void *do_client(void *arg)
{
	char *file_buff;
	int ret;
	thread_info_t *ti = (thread_info_t *)arg;
	char fname[256];

	sprintf(fname, "output.%d", ti->id);

	/*establish connection with server*/
	pthread_mutex_lock(&connect_mutex);
	make_connection(ti);
	pthread_mutex_unlock(&connect_mutex);

	rsf_out(5, "[%d] Connected\n", ti->id);

	rsf_out(5, "[%d] Allocating a buffer of %lu bytes\n", ti->id, bufsize);
	file_buff = malloc(bufsize);

	if (ti->transport_mode == TCP) {  /*usual TCP mode*/
		rsf_out(5, "[%d] Getting file\n", ti->id);
		int i, offset = 0;
		int num = bufsize;

		while (num > 0) {
			errno = 0;
			i = read(ti->conn_fd, file_buff + offset, num);
			if (i < 0 && errno != EAGAIN) {
				perror("read");
				pthread_exit(0);
			}
			if (errno == EAGAIN) {
				continue;
			}
			if (i == 0) {
				continue;
			}
			num -= i;
			offset += i;

		}
	#ifdef WRITE_FILE
		ret = open(fname, O_CREAT | O_TRUNC | O_WRONLY, 0666);
		if(ret < 0){
			rsf_out(0, "Can't open output file\n");
		}
		sock_write(ret, file_buff, bufsize);
	#endif

	} else {
		dest_sge_t my_sge;
		struct ibv_send_wr *bad_wr;

		rsf_out(5, "[%d] Using RDMA to get file\n", ti->id);
		/*Send server my buffer info*/
		my_sge.to = uint64_from_ptr(ti->rdma_buf);
		my_sge.stag = ti->rdma_mr->rkey;
		my_sge.length = bufsize;
		rsf_out(5, "[%d] RDMA Address = 0x%lx RDMA Rkey = %d Length is %d\n",
			ti->id, my_sge.to, my_sge.stag, my_sge.length);
		memcpy(ti->send_buf, &my_sge, sizeof(my_sge));

		memset(ti->rdma_buf, '\0', bufsize);
		memset(ti->recv_buf, '\0', control_size);


		#ifdef BLOCKING
		if (ti->cq_channel != NULL) {
			rsf_out(5, "Requesting notification of next post send\n");
			ret = ibv_req_notify_cq(ti->cq_hndl, 0);
			if(ret){
				rsf_out(0, "Unable to request notification on CQ\n");
				pthread_exit(0);
			}
		}
		#endif


		ret = ibv_post_send(ti->qp_hndl, &ti->send_wr, &bad_wr);
		if (ret) {
			iwarp_error("Unable to post send", ret);
			pthread_exit(0);
		}


		/*poll the send off the completion queue*/
		poll_cq_block_threaded(ti);


		#ifdef BLOCKING
		if (ti->cq_channel != NULL) {
			rsf_out(5, "Requesting notification of next post send\n");
			ret = ibv_req_notify_cq(ti->cq_hndl, 0);
			if(ret){
				rsf_out(0, "Unable to request notification on CQ\n");
				pthread_exit(0);
			}
		}
		#endif


		/*wait for done message*/
		poll_cq_block_threaded(ti);

	#ifdef WRITE_FILE
		ret = open(fname, O_CREAT | O_TRUNC | O_WRONLY, 0666);
		if(ret < 0){
			iwarp_error(0, "Can't open output file\n");
		}
		sock_write(ret, ti->rdma_buf, bufsize);
	#endif

		ret = rdma_disconnect(ti->cma_id);  /*only one side does this*/
		if (ret ) {
			iwarp_error("Unable to disconnect RDMA", ret);
		}


	}

	close_threaded_connection(ti);





	pthread_exit(0);
}

static void send_done(thread_info_t *ti)
{
	int ret;
	struct ibv_send_wr *bad_wr;

	if (ti->transport_mode == TCP) {
		return;
	}

	#ifdef BLOCKING
	if (ti->cq_channel != NULL) {
		rsf_out(5, "Requesting notification of next post send\n");
		ret = ibv_req_notify_cq(ti->cq_hndl, 0);
		if(ret){
			rsf_out(0, "Unable to request notification on CQ\n");
			pthread_exit(0);
		}
	}
	#endif

	/*now need to send the done message*/
	strcpy(ti->send_buf, "DONE");
	ret = ibv_post_send(ti->qp_hndl, &ti->send_wr, &bad_wr);
	if (ret) {
		iwarp_error("Unable to post send", ret);
		pthread_exit(0);
	}

	/*poll the send off the completion queue*/
	poll_cq_block_threaded(ti);
}

int main(int argc, char **argv)
{
	char *file_name = NULL;
	int am_server;

	transport_mode_t transport_mode = RDMA; /*rmda=0 tcp=1*/
	sendfile_mode_t sendfile_mode = SENDFILE;
	am_server = 1;  /*server by default*/
	thread_info_t *ti[256];
	pthread_t thread_list[256];
	pthread_attr_t attr;
	int i;
	int ret;

	pthread_mutex_init(&count_mutex, NULL);
	pthread_mutex_init(&client_done_mutex, NULL);
	pthread_mutex_init(&connect_mutex, NULL);
	pthread_cond_init(&client_done, NULL);

	count = 0;

	server = NULL;
	while (++argv, --argc > 0) {
		if (**argv == '-') {
			switch ((*argv)[1]) {
			case 'c':
				am_server = 0;
				argv++;
				argc--;
				if (argc == 0)
					usage();
				server = *argv;
				break;
			case 'f':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				file_name = *argv;
				break;
			case 'w':
				sendfile_mode = WRITE;
				break;
			case 't':
				transport_mode = TCP;
				break;
			case 's':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				bufsize = atol(*argv);
			        break;
			case 'h':
				usage();
				return -1;
			case 'p':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				PAGE_THRESHOLD = atoi(*argv);
				break;
			case 'o':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				MAX_OPS = atoi(*argv);
					break;
			case 'x':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				server = *argv;
				break;
			case 'z':
				argv++;
				argc--;
				if (argc == 0)
					usage();
				max_threads = atoi(*argv);
				break;
			}
		}
	}

	if (server == NULL) {
		rsf_out(0, "\n\n\nNeed to specify server IP because of buggy NES driver, use -x\n\n\n");
		usage();
	}


	pthread_attr_init(&attr);
	pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);

	if (am_server){
		rsf_out(1, "Threads= %d\n", max_threads);
		for (i=0; i<max_threads; i++) {
			ti[i] = malloc(sizeof (thread_info_t));
			ti[i]->transport_mode = transport_mode;
			ti[i]->sendfile_mode = sendfile_mode;
			ti[i]->filename = file_name;
			ti[i]->id = i;
			/*spawn thread have it run wait_connection*/
			ret = pthread_create(&thread_list[i], NULL,
					     wait_connection, ti[i]);

		}
		rsf_out(1, "Done spawning threads going to wait\n");

		/*main thread needs to wait for the rest to stop*/
		for (i=0; i<max_threads; i++) {
			pthread_join(thread_list[i], NULL);
			rsf_out(5, "Thread: %d finished\n", i);
		}

		printf("BufSize= %d Threads= %d Time= %f (s) BW= %f Mbps\n",
			(int)bufsize, max_threads,
			elapsed_wall_time(start, stop, SECONDS),
			(bufsize*max_threads)/1024/1024*8/elapsed_wall_time(start, stop, SECONDS));

	} else {
		rsf_out(1, "Active Threads= %d\n", max_threads);
		for (i=0; i<max_threads; i++) {
			ti[i] = malloc(sizeof (thread_info_t));
			ti[i]->transport_mode = transport_mode;
			ti[i]->sendfile_mode = sendfile_mode;
			ti[i]->id = i;
			ret = pthread_create(&thread_list[i], &attr,
				do_client, ti[i]);
		}
	}
	pthread_attr_destroy(&attr);
	pthread_mutex_destroy(&client_done_mutex);
	pthread_mutex_destroy(&count_mutex);
	pthread_mutex_destroy(&connect_mutex);
	pthread_cond_destroy(&client_done);


	pthread_exit(0);

	return 0;
}

